import java.io.*;
import javax.swing.tree.DefaultMutableTreeNode;
import java.util.Enumeration;
import java.util.Vector;
import java.util.Arrays;
import java.util.List;
import java.util.ArrayList;
import java.lang.Math;
import javax.swing.*;

public class AlgoApp
{
	double SIM_THRESHOLD = 0.2; // threshold to be reached to accept a similarity between 2 nodes
	double REPLACE_THRESHOLD = 0.2; // threshold to be reached to replace strings
	int nbmappings =0;
	String[] filePaths;
	SpecialTreeNode[] schemaRoots;
	BufferedWriter bw;
	
	public AlgoApp(String[] paths)
	{
		filePaths = paths;
		schemaRoots = new SpecialTreeNode [filePaths.length];
		try
		{ bw = new BufferedWriter(new FileWriter("approxivectMappings.txt")); }
		catch(IOException ioe) {System.out.println(ioe.toString());}
	}

	public void printTrees()
	{
	    /*JTree tree = new JTree(rootSource);
		JScrollPane treeView = new JScrollPane(tree);
		JFrame frame1 = new JFrame("pfff");
		frame1.add(treeView);
		frame1.pack();
		frame1.setVisible(true);
		JTree tree2 = new JTree(rootTarget);
		JScrollPane treeView2 = new JScrollPane(tree2);
		JFrame frame2 = new JFrame("pfff");
		frame2.add(treeView2);
		frame2.pack();
		frame2.setVisible(true);
	    */
	    for(int i=0; i<schemaRoots.length; i++) // matcher les schmas 2 par 2
		{
	    	Enumeration e = schemaRoots[i].breadthFirstEnumeration();
	    	SpecialTreeNode stn;
		    while(e.hasMoreElements())
		    {
		        stn = (SpecialTreeNode)e.nextElement();
				System.out.println(stn.getOriginalString()+" ::: "+stn.getLevel());
			}
		}
	}

	public void runAlgo(SpecialTreeNode rootSource, SpecialTreeNode rootTarget) throws IOException
	{
		Vector<SpecialTreeNode> tempVect1, tempVect2;
		SpecialTreeNode tempNode1, tempNode2;
		Enumeration e1 = rootSource.breadthFirstEnumeration();
		Enumeration e2;

		// step 1 : tree traversal for string replacement
		while(e1.hasMoreElements())
		{
		    tempNode1 = (SpecialTreeNode) e1.nextElement();
			tempVect1 = tempNode1.getChildren();
			e2 = rootTarget.breadthFirstEnumeration();
			while(e2.hasMoreElements())
			{
		    	tempNode2 = (SpecialTreeNode) e2.nextElement();
		    	tempVect2 = tempNode2.getChildren();
		    	compareVectors(tempVect1, tempVect2);
		    }
		}
		
		e1 = rootSource.breadthFirstEnumeration();
		// step 2 : tree traversal for cosine measure
		while(e1.hasMoreElements())
		{
		    tempNode1 = (SpecialTreeNode) e1.nextElement();
			tempVect1 = tempNode1.Neighbourhood();
			e2 = rootTarget.breadthFirstEnumeration();
			while(e2.hasMoreElements())
			{
		    	tempNode2 = (SpecialTreeNode) e2.nextElement();
		    	tempVect2 = tempNode2.Neighbourhood();
			    // prise en compte de la structure : on mappe que des feuilles ensemble ou des non feuilles ensemble
        		if((tempNode1.isLeaf() && tempNode2.isLeaf()) || (!tempNode2.isLeaf() && !tempNode1.isLeaf()))
        		{
        		    String st1 = tempNode1.getOriginalString();
					String st2 = tempNode2.getOriginalString();
			    	double cos = mesureCosinus(tempVect1, tempVect2);
			    	double str = mesureString(st1, st2);
					//double res = max(cos, str);
					double res = (cos+str)/2;
					if(res >= SIM_THRESHOLD)
					{
					    String s = st1+";"+st2+";"+res;
					    //System.out.println("SIMILARITY : "+s);
						bw.write(s, 0, s.length());
						bw.newLine();
						nbmappings++;
					}
				}
		    }
		}
	}

	void compareVectors(Vector<SpecialTreeNode> v1, Vector<SpecialTreeNode> v2)
	{
	    // replace string from one vector if similarity with another smaller-sized string.
	    SpecialTreeNode tempNode1, tempNode2;
	    for(int i=0;i<v1.size();i++)
        {
            tempNode1 = (SpecialTreeNode) v1.get(i);
            String st1 = (String) tempNode1.getUserObject();
			for(int j=0;j<v2.size();j++)
			{
			    tempNode2 = (SpecialTreeNode) v2.get(j);
			    // prise en compte de la structure : on mappe que des feuilles ensemble ou des non feuilles ensemble
        		if((tempNode1.isLeaf() && tempNode2.isLeaf()) || (!tempNode2.isLeaf() && !tempNode1.isLeaf()))
        		{
				    String st2 = (String) tempNode2.getUserObject();
					double sim = mesureString(st1, st2);
					if(sim >= REPLACE_THRESHOLD)
					{
						//System.out.println("replacing : "+st1+" + "+st2+" = "+sim);
					    if(st1.length() > st2.length()) ((SpecialTreeNode) v1.get(i)).setUserObject(st2);
					    else ((SpecialTreeNode) v2.get(j)).setUserObject(st1);
					}
				}
			}
        }
	}

	double mesureCosinus(Vector<SpecialTreeNode> vn1, Vector<SpecialTreeNode> vn2)
	{
	    Vector<SpecialTreeNode> v1 = new Vector<SpecialTreeNode>(vn1);
	    Vector<SpecialTreeNode> v2 = new Vector<SpecialTreeNode>(vn2);
	    Vector<Double> newVect1 = new Vector<Double>();
	    Vector<Double> newVect2 = new Vector<Double>();
	    SpecialTreeNode tempNode1, tempNode2;
	    String st1, st2;
	    double nb;

	    while(!(v1.isEmpty()))
	    {
	        tempNode1 = (SpecialTreeNode) v1.remove(0); // get 1st node in the vector
	        st1 = (String) tempNode1.getUserObject();
	        nb=1*tempNode1.getPoids();
	        for(int i=0; i<v1.size();i++)
	        {
	            tempNode2 = (SpecialTreeNode) v1.get(i);
	            st2 = (String) tempNode2.getUserObject();
	            if(st1.equals(st2))
	            {
	                nb+=(1*tempNode2.getPoids());
	                v1.remove(i);
	            }
	        }
	        newVect1.add((double)(nb));
	        nb=0;
	        for(int i=0; i<v2.size();i++)
	        {
	            tempNode2 = (SpecialTreeNode) v2.get(i);
	            st2 = (String) tempNode2.getUserObject();
	            if(st1.equals(st2))
	            {
	                nb+=1*tempNode2.getPoids();
	                v2.remove(i);
	            }
	        }
			newVect2.add(nb);
	    }
	    while(!(v2.isEmpty()))
	    {

	        tempNode1 = (SpecialTreeNode) v2.remove(0); // get 1st node in the vector
	        st1 = (String) tempNode1.getUserObject();
	        nb=1*tempNode1.getPoids();
	        for(int i=0; i<v2.size();i++)
	        {
	            tempNode2 = (SpecialTreeNode) v2.get(i);
	            st2 = (String) tempNode2.getUserObject();
	            if(st1.equals(st2))
	            {
	                nb+=1*tempNode2.getPoids();
	                v2.remove(i);
	            }
	        }
	        newVect1.add(0.0); // vecteur v1 empty so no occurence of st2 in v1
			newVect2.add(nb);
	    }

	    int num = 0;
	    double denom1 = 0.0;
	    double denom2 = 0.0;

	    for(int i=0;i<newVect1.size();i++)
	    {
	        num += (double)newVect1.get(i)*(double)newVect2.get(i);
			denom1 += (double)newVect1.get(i)*(double)newVect1.get(i);
			denom2 += (double)newVect2.get(i)*(double)newVect2.get(i);
	    }
	    denom1 = (double)Math.sqrt(denom1);
	    denom2 = (double)Math.sqrt(denom2);
		return (double)(num/((double)denom1*denom2));
	}

	public void init()
	{
        XMLParser xmlpars = new XMLParser();
		for(int i=0; i<filePaths.length; i++) // parser les schemas
		{
        	schemaRoots[i] = new SpecialTreeNode(filePaths[i]);
        	xmlpars.startParsing(filePaths[i], schemaRoots[i]);
			schemaRoots[i] = schemaRoots[i].getChild(0); // on supprime le schema element
		}
		for(int i=0; i<schemaRoots.length; i++) // matcher les schmas 2 par 2
		{
		    for(int j=i+1 ; j<schemaRoots.length; j++)
		    {
		    	try { runAlgo(schemaRoots[i], schemaRoots[j]); }
				catch(IOException ioe) {System.out.println(ioe.toString());}
			}
		}
		//printTrees();
		try{ bw.close(); } catch(IOException ioe) {ioe.printStackTrace();}
		System.out.println("fini : "+nbmappings+" found mappings on "+schemaRoots.length+" schemas");
	}
	
	public double triGrammes(String s1, String s2)
	{
	    double res = 0;
	    Vector<String> triS1 = new Vector<String>();
		Vector<String> triS2 = new Vector<String>();
	    int compteur=0;
	    
	    while(s1.substring(compteur).length()>2)
		{
			triS1.add(s1.substring(compteur, compteur+3));
			compteur++;
		}
		compteur = 0;
		while(s2.substring(compteur).length()>2)
		{
			triS2.add(s2.substring(compteur, compteur+3));
			compteur++;
		}
		
		int nbCommonTriGr = 0;
		for(int i=0;i<triS1.size();i++)
		    for(int j=0;j<triS2.size();j++)
		    {
		        if(triS1.get(i).equals(triS2.get(j))) nbCommonTriGr++;
		    }

		res = (1.0 / (1.0 + (double)triS1.size() + (double)triS2.size() - (double)(2*nbCommonTriGr)));
		return res;
	}
	
	public double levenshteinDistance(String s1, String s2)
	{
	    int cost, temp;
	    int matrix[][] = new int[s1.length()][s2.length()];
	    
	    for(int i=0;i<s1.length();i++) matrix[i][0]=i;
	    for(int j=0;j<s2.length();j++) matrix[0][j]=j;
	    for(int i=1;i<s1.length();i++)
	        for(int j=1;j<s2.length();j++)
	        {
	            if(s1.charAt(i)==s2.charAt(j)) cost = 0;
	            else cost = 1;
	            if(matrix[i-1][j] < matrix[i][j-1]) temp = matrix[i-1][j] + 1;
	            else temp = matrix[i][j-1] + 1;
	            if((matrix[i-1][j-1] + cost) < temp) temp = (matrix[i-1][j-1] + cost);
	            matrix[i][j] = temp;
	        }
	    
	    double minString;
	    if(s1.length() >= s2.length()) minString = (double)s2.length();
	    else minString = (double)s1.length();
	    double res = (minString - (double)matrix[s1.length()-1][s2.length()-1]) / minString;
		if(res < 0.0) res = 0;
	    return res;
	}
	
	public double jaccardSim(String s1, String s2)
	{
	    double res=0.0;
	    List<String> list1 = Arrays.asList(s1.split(" "));
	    List<String> list2 = Arrays.asList(s2.split(" "));
	    if(list1.size() > 1 || list2.size() > 1) // au moins 2 mots dans une des listes
	    {
	    	int totalOcc = list1.size() + list2.size();
	    	int commonOcc = 0;
	    	for(int i=0;i<list1.size();i++)
	    	{
	    	    String s = (String)list1.get(i);
	    	    if(list2.contains(s)) commonOcc++;
	    	}
	    	res = (double)commonOcc /((double)totalOcc - (double)commonOcc);
	    }
	    return res;
	}

	public double max(double n1, double n2)
	{if(n1>n2) return n1; else return n2;}

	public double mesureString(String s1, String s2)
	{
		double lev = levenshteinDistance(s1, s2);
	    double trig = triGrammes(s1, s2);
	    //double jacc = jaccardSim(s1, s2);
	    //System.out.println(s1+" / "+s2+" = "+lev+" + "+trig+" + "+jacc);
	    
	    // isobarycentre ou max entre la mesure de tri-grammes et le String Matching
	    //double res = (lev + trig)/2;
	    double res = max(lev, trig);
	    return res;
	}
}
