import javax.swing.tree.MutableTreeNode;
import javax.swing.tree.TreeNode;
import java.util.Vector;
import java.util.NoSuchElementException;
import java.util.Enumeration;

public class SpecialTreeNode implements MutableTreeNode
{
	static int NB_NIVEAUX = 1; // nb levels to go up and down to get the neighbours
	static double MIN_WEIGHT = 1.25; // minimum weight to accept a node as a neighbour
	static double K = 1.0;  // coefficient dans la formule de poids

	private String originalString;
	private String currentString;
	private SpecialTreeNode parent;
	private Vector<SpecialTreeNode> children;
	private double poids = 0.0;
	
	public SpecialTreeNode(String s)
	{
	    setUserObject(s);
	    originalString = s;
	    parent = null;
	    children = new Vector<SpecialTreeNode>();
	}

	public void removeFromParent() {}
	public void	remove(MutableTreeNode node) {}
	public void remove(int index) {}
	public boolean getAllowsChildren() {return true;}
	public SpecialTreeNode getChildAt(int childIndex) {return null;}
	public int getChildCount() {return children.size();}
	public int getIndex(TreeNode node) {return -1;}
	public void insert(MutableTreeNode child, int index){}
	public void setParent(MutableTreeNode par) {parent=(SpecialTreeNode)par;}
	public String getOriginalString() {return originalString;}
	public void setUserObject(Object o)	{currentString = (String) o;}
	public Object getUserObject() {return currentString;}
	public double getPoids() {return poids;}
	
	public Vector<SpecialTreeNode> getChildren() {return children;}
	public SpecialTreeNode getChild(int i) {return children.get(i);}
	public Enumeration<SpecialTreeNode> children(){return children.elements();}
	public void add(SpecialTreeNode c) {children.add(c); c.setParent(this);}
	public boolean isLeaf() {if(children.size()==0) return true; else return false;}
	public SpecialTreeNode getParent() {return parent;}
	
	public int getLevel()
	{
	    if(parent==null) return 0;
	    else return 1+parent.getLevel();
	}

	double calculateWeight(SpecialTreeNode n)
	{
	    // weight of this according to current node n
	    if(this == n) return K+1.0;
	    double weight=0.0;
	    SpecialTreeNode commonAncestor;
	    
	    if(isNodeDescendant(n)) commonAncestor = this;
        else if(isNodeAncestor(n)) commonAncestor = n;
        else commonAncestor = getSharedAncestor(n);
		
	    if(commonAncestor == n || commonAncestor == this) weight = (double)(Math.abs(this.getLevel() - n.getLevel()) + Math.abs(this.getLevel() - commonAncestor.getLevel()) + Math.abs(n.getLevel() - commonAncestor.getLevel()));
	    else weight = (double)(2 * (Math.abs(this.getLevel() - commonAncestor.getLevel()) + Math.abs(n.getLevel() - commonAncestor.getLevel())));
		this.poids = 1.0+(K / weight);
	    return this.poids;
	}

	public Vector<SpecialTreeNode> Neighbourhood()
	{
	    // returns the neighbourhood of this node with their respective weight
		Vector<SpecialTreeNode> retVect = new Vector<SpecialTreeNode>();
		SpecialTreeNode ancetre = this;
		if(NB_NIVEAUX == 0) retVect.add(ancetre);
		// cas racine
		else if(ancetre.getParent()==null)
		{
		    retVect.add(ancetre);
			retVect.addAll(ancetre.getDescendantNeighbours(NB_NIVEAUX, this));
		}
		else
		{
		    int endUp = NB_NIVEAUX;
		    int levelMax = NB_NIVEAUX + this.getLevel();
			// going up to the right ancestor
			while((ancetre.getParent()) != null && endUp > 0)
			{
			    ancetre = ancetre.getParent();
			    endUp--;
			}
			// going down to all descendents
			retVect.add(ancetre);
		    retVect.addAll(ancetre.getDescendantNeighbours(levelMax, this));
		}
		return retVect;
	}

    public Vector<SpecialTreeNode> getDescendantNeighbours(int levelMax, SpecialTreeNode currentNode)
    {
		double poids;
    	Vector<SpecialTreeNode> retVect = new Vector<SpecialTreeNode>();
		if(this.getLevel() < levelMax)
		{
			for(int i=0;i<children.size();i++)
			{
			    SpecialTreeNode tempNode = (SpecialTreeNode) children.get(i);
			    tempNode.calculateWeight(currentNode);
			    if(tempNode.getPoids() >= MIN_WEIGHT) retVect.add(tempNode);
			    retVect.addAll(tempNode.getDescendantNeighbours(levelMax, currentNode));
			}
		}
		return retVect;
    }
    


//////////////////////////////////////////
////// fonctions DefaultMutableTreeNode
//////////////////////////////////////////

	
	public Enumeration<SpecialTreeNode> breadthFirstEnumeration()
	{return new BreadthFirstEnumeration(this);}
	
    public boolean isNodeAncestor(SpecialTreeNode anotherNode)
	{
		if (anotherNode == null) return false;
		SpecialTreeNode ancestor = this;
		do
		{
		    if (ancestor == anotherNode) return true;
		} while((ancestor = ancestor.getParent()) != null);
		return false;
    }

    public boolean isNodeDescendant(SpecialTreeNode anotherNode)
	{
		if (anotherNode == null) return false;
		return anotherNode.isNodeAncestor(this);
    }
    
    public SpecialTreeNode getSharedAncestor(SpecialTreeNode aNode)
	{
		if (aNode == this) return this;
		else if (aNode == null) return null;

		int	level1, level2, diff;
		SpecialTreeNode node1, node2;

		level1 = getLevel();
		level2 = aNode.getLevel();

		if (level2 > level1)
		{
		    diff = level2 - level1;
		    node1 = aNode;
		    node2 = this;
		}
		else
		{
		    diff = level1 - level2;
		    node1 = this;
		    node2 = aNode;
		}

		// Go up the tree until the nodes are at the same level
		while (diff > 0) {
		    node1 = node1.getParent();
		    diff--;
		}

		// Move up the tree until we find a common ancestor.  Since we know
		// that both nodes are at the same level, we won't cross paths
		// unknowingly (if there is a common ancestor, both nodes hit it in
		// the same iteration).
		do {
		    if (node1 == node2)	return node1;
		    node1 = node1.getParent();
		    node2 = node2.getParent();
		} while (node1 != null);// only need to check one -- they're at the
		// same level so if one is null, the other is

		if (node1 != null || node2 != null) throw new Error ("nodes should be null");
		return null;
    }
    
    // BEGINNING CLASS BreadthFirstEnumeration
    final class BreadthFirstEnumeration implements Enumeration<SpecialTreeNode>
	{
		protected Queue	queue;

		public BreadthFirstEnumeration(SpecialTreeNode rootNode) {
		    super();
		    Vector<SpecialTreeNode> v = new Vector<SpecialTreeNode>(1);
		    v.addElement(rootNode);
		    queue = new Queue();
		    queue.enqueue(v.elements());
		}

		public boolean hasMoreElements() {
		    return (!queue.isEmpty() &&
			    ((Enumeration)queue.firstObject()).hasMoreElements());
		}

		public SpecialTreeNode nextElement() {
		    Enumeration	enumer = (Enumeration)queue.firstObject();
		    SpecialTreeNode	node = (SpecialTreeNode)enumer.nextElement();
		    Enumeration	children = node.children();

		    if (!enumer.hasMoreElements()) {
			queue.dequeue();
		    }
		    if (children.hasMoreElements()) {
			queue.enqueue(children);
		    }
		    return node;
		}
	
		// A simple queue with a linked list data structure.
		final class Queue {
		    QNode head;	// null if empty
		    QNode tail;

		    final class QNode {
			public Object	object;
			public QNode	next;	// null if end
			public QNode(Object object, QNode next) {
			    this.object = object;
			    this.next = next;
			}
		    }

		    public void enqueue(Object anObject) {
			if (head == null) {
			    head = tail = new QNode(anObject, null);
			} else {
			    tail.next = new QNode(anObject, null);
			    tail = tail.next;
			}
		    }

		    public Object dequeue() {
			if (head == null) {
			    throw new NoSuchElementException("No more elements");
			}

			Object retval = head.object;
			QNode oldHead = head;
			head = head.next;
			if (head == null) {
			    tail = null;
			} else {
			    oldHead.next = null;
			}
			return retval;
		    }

		    public Object firstObject() {
			if (head == null) {
			    throw new NoSuchElementException("No more elements");
			}

			return head.object;
		    }

		    public boolean isEmpty() {
			return head == null;
		    }

		} // End of class Queue
    }  // End of class BreadthFirstEnumeration
    

}
